//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"database/sql"
	"fmt"
	"net/http"
	"regexp"
	"strconv"
	"strings"

	"humungus.tedunangst.com/r/webs/login"
)

type Filter struct {
	ID      int64 `json:"-"`
	userid  UserID
	Name    string        `json:",omitempty"`
	Notes   string        `json:",omitempty"`
	Type    string        `json:",omitempty"`
	Matcher *filtMatcher  `json:",omitempty"`
	Actions []*filtAction `json:",omitempty"`
}

type filtMatcher struct {
	OP     string `json:",omitempty"`
	Val    string `json:",omitempty"`
	re     *regexp.Regexp
	Childs []*filtMatcher `json:",omitempty"`
}

func (matcher *filtMatcher) String() string {
	switch matcher.OP {
	case "any":
		return "any"
	case "and":
		var buf []string
		for _, m := range matcher.Childs {
			buf = append(buf, m.String())
		}
		return fmt.Sprintf("(%s)", strings.Join(buf, " "))
	}
	return fmt.Sprintf("%s:%s", matcher.OP, matcher.Val)
}

type filtAction struct {
	OP string
}

type filtResult int

const (
	filtResNone     filtResult = 0
	filtResHide     filtResult = 1
	filtResReject   filtResult = 2
	filtResNoImages filtResult = 4
)

func (res filtResult) Reject() bool {
	return res&filtResReject == filtResReject
}
func (res filtResult) NoImages() bool {
	return res&filtResNoImages == filtResNoImages
}

func (action *filtAction) String() string {
	return fmt.Sprintf("%s", action.OP)
}

var allfilters = make(map[string][]*Filter)

func filterPost(post *Post) filtResult {
	filts := allfilters["(site):post"]
	var res filtResult
	for _, filt := range filts {
		res |= filt.Apply(post)
	}
	return res
}

func filterComment(comment *Comment) filtResult {
	filts := allfilters["(site):comment"]
	var res filtResult
	for _, filt := range filts {
		res |= filt.Apply(comment)
	}
	return res
}

func filterPostsForUser(user *User, posts []*Post) []*Post {
	filts := allfilters[user.Name+":post"]
	if len(filts) == 0 {
		return posts
	}
	res := make([]*Post, 0, len(posts))
nextpost:
	for _, post := range posts {
		for _, filt := range filts {
			rv := filt.Apply(post)
			if rv == filtResHide {
				continue nextpost
			}
		}
		res = append(res, post)
	}
	return res
}

func filterCommentsForUser(user *User, comments []*Comment) []*Comment {
	filts := allfilters[user.Name+":comment"]
	if len(filts) == 0 {
		return comments
	}
	res := make([]*Comment, 0, len(comments))
	var hidden map[int64]bool
nextcomment:
	for _, comment := range comments {
		if hidden != nil && hidden[comment.ParentID] {
			hidden[comment.ID] = true
			continue
		}
		for _, filt := range filts {
			rv := filt.Apply(comment)
			if rv == filtResHide {
				if hidden == nil {
					hidden = make(map[int64]bool)
				}
				hidden[comment.ID] = true
				continue nextcomment
			}
		}
		res = append(res, comment)
	}
	return res
}

type filterable interface {
	filtMatch(*filtMatcher) bool
	filtApply(*filtAction) filtResult
}

func (filt *Filter) Apply(what filterable) filtResult {
	if !filt.Matcher.Match(what) {
		return filtResNone
	}
	for _, action := range filt.Actions {
		rv := action.Apply(what)
		if rv != filtResNone {
			return rv
		}
	}
	return filtResNone
}

func (matcher *filtMatcher) Match(what filterable) bool {
	switch matcher.OP {
	case "any":
		return true
	case "and":
		ok := true
		for _, m := range matcher.Childs {
			ok = ok && m.Match(what)
		}
		return ok
	}
	return what.filtMatch(matcher)
}

func (what *Post) filtMatch(matcher *filtMatcher) bool {
	switch matcher.OP {
	case "from":
		return matcher.Val == what.PersonName || matcher.Val == originate(what.APid)
	case "newposter":
		return what.PersonID == 0
	case "group":
		return matcher.Val == what.GroupName
	case "link":
		return matcher.re.MatchString(what.Link)
	case "site":
		return matcher.re.MatchString(linkhost(what.Link))
	case "title":
		return matcher.re.MatchString(what.Title)
	}
	return false

}

func (what *Comment) filtMatch(matcher *filtMatcher) bool {
	switch matcher.OP {
	case "from":
		return matcher.Val == what.PersonName || matcher.Val == originate(what.APid)
	}
	return false
}

var re_noimg = regexp.MustCompile("<img [^>]*>")

func (action *filtAction) Apply(what filterable) filtResult {
	switch action.OP {
	case "hide":
		return filtResHide
	case "reject":
		return filtResReject
	case "nop":
		return filtResNone
	}
	return what.filtApply(action)
}

func (what *Post) filtApply(action *filtAction) filtResult {
	switch action.OP {
	case "relink":
		if link := findlink(what.Content); link != "" {
			what.Link = normalizeLink(link, originate(what.APid))
			what.Site = linkhost(what.Link)
		}
	case "noimg":
		what.Content = re_noimg.ReplaceAllString(what.Content, "")
	}
	return filtResNone
}

func (what *Comment) filtApply(action *filtAction) filtResult {
	switch action.OP {
	case "noimg":
		what.Content = re_noimg.ReplaceAllString(what.Content, "")
	}
	return filtResNone
}

func (filt *Filter) compile() error {
	if err := filt.Matcher.compile(); err != nil {
		return err
	}
	for _, a := range filt.Actions {
		if err := a.compile(); err != nil {
			return err
		}
	}
	return nil
}

func (matcher *filtMatcher) compile() error {
	for _, m := range matcher.Childs {
		if err := m.compile(); err != nil {
			return err
		}
	}
	var err error
	switch matcher.OP {
	case "link":
		fallthrough
	case "site":
		fallthrough
	case "title":
		matcher.re, err = regexp.Compile(matcher.Val)
	}
	return err
}

func (action *filtAction) compile() error {
	return nil
}

func showadminfilters(w http.ResponseWriter, r *http.Request) {
	filters := allfilters["(site):post"]
	filters = append(filters, allfilters["(site):comment"]...)
	templinfo := getInfo(r)
	templinfo["SaveAction"] = "/saveadmin"
	templinfo["Filters"] = filters
	templinfo["SaveCSRF"] = login.GetCSRF("saveadmin", r)
	err := readviews.Execute(w, "adminfilters.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showuserfilters(w http.ResponseWriter, r *http.Request) {
	u := login.GetUserInfo(r)
	user := finduser(u.Username)
	filters := allfilters[user.Name+":post"]
	filters = append(filters, allfilters[user.Name+":comment"]...)
	templinfo := getInfo(r)
	templinfo["SaveAction"] = "/saveaccount"
	templinfo["Filters"] = filters
	templinfo["SaveCSRF"] = login.GetCSRF("saveaccount", r)
	err := readviews.Execute(w, "accountfilters.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func parseMatches(input string) *filtMatcher {
	strs := strings.Split(input, " ")
	var matches []*filtMatcher
	for _, s := range strs {
		v := strings.SplitN(s, ":", 2)
		if len(v) == 1 {
			v = append(v, "")
		}
		m := new(filtMatcher)
		m.OP = v[0]
		switch m.OP {
		case "any":
		case "newposter":
		case "from":
			m.Val = v[1]
		case "group":
			m.Val = v[1]
		case "link":
			m.Val = v[1]
			_, err := regexp.Compile(m.Val)
			if err != nil {
				return nil
			}
		case "site":
			m.Val = v[1]
			_, err := regexp.Compile(m.Val)
			if err != nil {
				return nil
			}
		case "title":
			m.Val = v[1]
			_, err := regexp.Compile(m.Val)
			if err != nil {
				return nil
			}
		default:
			return nil
		}
		matches = append(matches, m)
	}
	if len(matches) == 0 {
		return &filtMatcher{OP: "any"}
	}
	if len(matches) > 1 {
		return &filtMatcher{OP: "and", Childs: matches}
	}
	return matches[0]
}

func parseActions(input string) []*filtAction {
	strs := strings.Split(input, " ")
	var actions []*filtAction
	for _, s := range strs {
		v := strings.SplitN(s, ":", 2)
		if len(v) == 1 {
			v = append(v, "")
		}
		a := new(filtAction)
		a.OP = v[0]
		switch a.OP {
		case "hide":
		case "reject":
		case "relink":
		case "noimg":
		case "nop":
		default:
			return nil
		}
		actions = append(actions, a)
	}
	return actions
}

func savefilter(userid UserID, w http.ResponseWriter, r *http.Request) bool {
	isDelete := r.FormValue("delete") == "1"
	if isDelete {
		filterid, _ := strconv.ParseInt(r.FormValue("filterid"), 10, 0)
		_, err := stmtDeleteFilter.Exec(userid, filterid)
		if err != nil {
			elog.Printf("error deleting filter: %s", err)
		}
		loadFilters()
		return true
	}
	name := r.FormValue("name")
	notes := r.FormValue("notes")
	typ := r.FormValue("type")
	if typ != "post" && typ != "comment" {
		errorpage(w, r, "type must be page or comment")
		return false
	}
	matcher := parseMatches(r.FormValue("match"))
	actions := parseActions(r.FormValue("action"))
	if matcher == nil {
		errorpage(w, r, "error parsing match")
		return false
	}
	if actions == nil {
		errorpage(w, r, "error parsing actions")
		return false
	}
	filter := &Filter{
		userid:  userid,
		Name:    name,
		Notes:   notes,
		Type:    typ,
		Matcher: matcher,
		Actions: actions,
	}
	err := dbSaveFilter(filter)
	if err != nil {
		elog.Printf("error saving filter: %s", err)
	}
	loadFilters()
	return true
}

func saveadminfilters(w http.ResponseWriter, r *http.Request) {
	ok := savefilter(0, w, r)
	if !ok {
		return
	}
	http.Redirect(w, r, "/admin/filters", http.StatusSeeOther)
}

func websaveuserfilters(w http.ResponseWriter, r *http.Request) {
	u := login.GetUserInfo(r)
	ok := savefilter(UserID(u.UserID), w, r)
	if !ok {
		return
	}
	http.Redirect(w, r, "/account/filters", http.StatusSeeOther)
}

func dbSaveFilter(filter *Filter) error {
	j, err := jsonify(filter)
	if err == nil {
		if filter.ID == 0 {
			var res sql.Result
			res, err = stmtSaveFilter.Exec(filter.userid, j)
			if err == nil {
				filter.ID, _ = res.LastInsertId()
			}
		} else {
			panic("no")
		}
	}
	return err
}

func loadFilters() {
	rows, err := stmtGetFilters.Query()
	if err != nil {
		elog.Printf("error querying filters: %s", err)
		return
	}
	defer rows.Close()
	var filters []*Filter
	for rows.Next() {
		filt := new(Filter)
		var j string
		err = rows.Scan(&filt.ID, &filt.userid, &j)
		if err == nil {
			err = unjsonify(j, filt)
		}
		if err == nil {
			err = filt.compile()
		}
		if err != nil {
			elog.Printf("error scanning filter: %s", err)
			continue
		}
		filters = append(filters, filt)
	}
	m := make(map[string][]*Filter)
	for _, filt := range filters {
		var name string
		if filt.userid == 0 {
			name = "(site)"
		} else {
			user := dbGetUser(filt.userid)
			if user == nil {
				continue
			}
			name = user.Name
		}
		name += ":" + filt.Type
		m[name] = append(m[name], filt)
	}
	allfilters = m
}

func threadcomments(comments []*Comment) []*Comment {
	thread := make([]*Comment, 0, len(comments)/2)
	idmap := make(map[int64]*Comment, len(comments))
	for _, c := range comments {
		idmap[c.ID] = c
	}
	for _, c := range comments {
		arr := &thread
		if p := idmap[c.ParentID]; p != nil {
			arr = &p.Kids
		}
		if l := len(*arr); l > 0 {
			(*arr)[l-1].Next = c
		}
		*arr = append(*arr, c)
	}
	return thread
}
