//
// Copyright (c) 2019 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"crypto/rand"
	"crypto/rsa"
	"database/sql"
	"fmt"
	"os"
	"regexp"
	"strings"

	"golang.org/x/crypto/bcrypt"
	_ "humungus.tedunangst.com/r/go-sqlite3"
	"humungus.tedunangst.com/r/termvc"
	"humungus.tedunangst.com/r/webs/httpsig"
	"humungus.tedunangst.com/r/webs/login"
)

var re_plainname = regexp.MustCompile("^[[:alnum:]_-]+$")

var alreadyopendb *sql.DB
var stmtConfig *sql.Stmt

func initdb() {
	os.Mkdir(dataDir+"/attachments", 0755)
	os.Mkdir(dataDir+"/tempstore", 0755)
	dbname := dataDir + "/azorius.db"
	_, err := os.Stat(dbname)
	if err == nil {
		elog.Fatalf("%s already exists", dbname)
	}
	db, err := sql.Open("sqlite3", dbname)
	if err != nil {
		elog.Fatal(err)
	}
	alreadyopendb = db

	cleanup := func() {
		os.Remove(dbname)
		os.Exit(1)
	}
	defer cleanup()

	_, err = db.Exec("PRAGMA journal_mode=WAL")
	if err != nil {
		elog.Print(err)
		return
	}
	for _, line := range strings.Split(sqlSchema, ";") {
		_, err = db.Exec(line)
		if err != nil {
			elog.Print(err)
			return
		}
	}

	prepareStatements(db)

	_, err = db.Exec("insert into persons values (-1, '[deleted]', '[deleted]', '.', '{}')")
	if err != nil {
		elog.Print(err)
		return
	}
	termvc.Start()
	defer cleanup()
	defer termvc.Restore()
	go termvc.Catch(cleanup)

	app := termvc.NewApp()
	t1 := termvc.NewTextArea()
	t1.Value = "\n\n\tHello.\n\t\tWelcome to azorius setup."
	var inputs []termvc.Element
	var offset int
	listenfield := termvc.NewTextInput("listen address", &offset)
	inputs = append(inputs, listenfield)
	serverfield := termvc.NewTextInput("server name", &offset)
	inputs = append(inputs, serverfield)
	namefield := termvc.NewTextInput("username", &offset)
	inputs = append(inputs, namefield)
	passfield := termvc.NewPasswordInput("password", &offset)
	inputs = append(inputs, passfield)
	okay := false
	btn := termvc.NewButton("let's go!")
	left := 25
	inputs = append(inputs, termvc.NewHPad(&left, btn, nil))
	form := termvc.NewForm(inputs...)
	t2 := termvc.NewTextArea()
	group := termvc.NewVStack(t1, form, t2)
	group.SetFocus(1)
	app.Element = group
	app.Screen = termvc.NewScreen()
	btn.Submit = func() {
		t2.Value = ""
		addr := listenfield.Value
		if len(addr) < 1 {
			t2.Value += "listen address is way too short.\n"
			elog.Print("that's way too short")
			return
		}
		setconfig("listenaddr", addr)
		addr = serverfield.Value
		if len(addr) < 1 {
			t2.Value += "server name is way too short.\n"
			return
		}
		setconfig("servername", addr)
		flags := flagIsApproved
		flags |= flagIsAdmin
		email := ""
		about := "what about me?"
		err = createuser(namefield.Value, passfield.Value, email, about, flags)
		if err != nil {
			t2.Value += fmt.Sprintf("error: %s\n", err)
			return
		}
		okay = true
		app.Quit()
	}

	app.Loop()
	if !okay {
		return
	}

	setcsrfkey()
	setconfig("dbversion", myVersion)

	setconfig("servermsg", "Things happen.")
	setconfig("aboutmsg", "### What is azorius?\nIt's pretty freaking alpha is what it is!")
	setconfig("loginmsg", "**login**")
	setconfig("devel", 0)

	db.Close()
	termvc.Restore()
	fmt.Printf("done.\n")
	os.Exit(0)
}

func setcsrfkey() {
	var randbytes [16]byte
	rand.Read(randbytes[:])
	key := fmt.Sprintf("%x", randbytes)
	setconfig("csrfkey", key)
}

func adduser() {
	panic("todo")
}

func chpass(username string) {
	panic("todo")
	user := finduser(username)
	if user == nil {
		elog.Fatal("no user")
	}
	db := opendatabase()
	login.Init(login.InitArgs{Db: db, Logger: ilog})

	pass := "password"
	err := login.SetPassword(int64(user.ID), pass)
	if err != nil {
		elog.Print(err)
		return
	}
	userid := user.ID
	_, err = db.Exec("delete from auth where userid = ?", userid)
	if err != nil {
		elog.Print(err)
		return
	}

	fmt.Printf("done\n")
	os.Exit(0)
}

func generateKeys() (string, string, error) {
	k, err := rsa.GenerateKey(rand.Reader, 2048)
	if err != nil {
		return "", "", err
	}
	pubkey, err := httpsig.EncodeKey(&k.PublicKey)
	if err != nil {
		return "", "", err
	}
	seckey, err := httpsig.EncodeKey(k)
	if err != nil {
		return "", "", err
	}
	return pubkey, seckey, nil
}

func createuser(name, pass, email, about string, flags int64) error {
	hash, err := bcrypt.GenerateFromPassword([]byte(pass), 12)
	if err != nil {
		return err
	}
	person := Person{
		Name:     name,
		Display:  name,
		About:    about,
		Format:   "markdown",
		Inbox:    fmt.Sprintf("https://%s/u/%s/inbox", serverName, name),
		ShareBox: fmt.Sprintf("https://%s/inbox", serverName),
	}
	err = dbSavePerson(&person)
	if err != nil {
		return err
	}
	apid := person.APid
	pubkey, seckey, err := generateKeys()
	if err == nil {
		err = dbSavePubkey(apid+"#key", apid, pubkey)
	}
	if err == nil {
		err = dbSaveSeckey(apid, seckey)
	}
	if err != nil {
		return err
	}
	user := User{
		Name:   name,
		Hash:   hash,
		Email:  email,
		Flags:  flags,
		Person: &person,
	}
	err = dbSaveUser(&user)
	if err != nil {
		return err
	}
	return nil
}

func opendatabase() *sql.DB {
	if alreadyopendb != nil {
		return alreadyopendb
	}
	dbname := dataDir + "/azorius.db"
	_, err := os.Stat(dbname)
	if os.IsNotExist(err) {
		elog.Fatalf("database %s does not exit", dbname)
	}
	if err != nil {
		elog.Fatalf("unable to open database: %s", err)
	}
	db, err := sql.Open("sqlite3", dbname)
	if err != nil {
		elog.Fatalf("unable to open database: %s", err)
	}
	stmtConfig, err = db.Prepare("select value from config where key = ?")
	if err != nil {
		elog.Fatal(err)
	}
	alreadyopendb = db
	return db
}

func getconfig(key string, value interface{}) error {
	m, ok := value.(*map[string]bool)
	if ok {
		rows, err := stmtConfig.Query(key)
		if err != nil {
			return err
		}
		defer rows.Close()
		for rows.Next() {
			var s string
			err = rows.Scan(&s)
			if err != nil {
				return err
			}
			(*m)[s] = true
		}
		return nil
	}
	row := stmtConfig.QueryRow(key)
	err := row.Scan(value)
	if err == sql.ErrNoRows {
		err = nil
	}
	return err
}

func setconfig(key string, val interface{}) error {
	db := opendatabase()
	db.Exec("delete from config where key = ?", key)
	_, err := db.Exec("insert into config (key, value) values (?, ?)", key, val)
	return err
}
